#! /usr/bin/Rscript

library(plyr)
library(reshape)
library(ggplot2)
library(robust)

# Fixed version of lmRob() per
# https://stat.ethz.ch/pipermail/r-sig-robust/2009/000285.html
lmRob = function (formula, data, weights, subset, na.action, model = TRUE,
    x = FALSE, y = FALSE, contrasts = NULL, nrep = NULL, control = lmRob.control(...),
    genetic.control = NULL, ...)
{
    the.call <- match.call()
    m <- match.call(expand = FALSE)
    m$model <- m$x <- m$y <- m$contrasts <- m$nrep <- NULL
    m$control <- m$genetic.control <- m$... <- NULL
    m[[1]] <- as.name("model.frame")
    m <- eval(m, sys.parent())
    Terms <- attr(m, "terms")
    weights <- model.extract(m, weights)
    Y <- model.extract(m, "response")
    X <- model.matrix(Terms, m, contrasts)
    if (nrow(X) <= ncol(X))
        stop("Robust method is inappropriate: not enough observations.")
    asgn <- attr(X, "assign")
    if (!is.list(asgn))
        asgn <- splus.assign(asgn, attr(Terms, "term.labels"))
    factor.names <- names(m)[sapply(m, is.factor)]
    at = attr(Terms,"factors")
    if (length(factor.names)>0)
      factor.names <- names(apply(at[ setdiff(dimnames(at)[[1]], factor.names),,
                                     drop=FALSE ], 2, sum)==0)
    x1.idx <- unname(unlist(asgn[factor.names]))
    X1 <- X[, x1.idx, drop = FALSE]
    X2 <- X[, setdiff(1:(dim(X)[2]), x1.idx), drop = FALSE]
    p2 <- ncol(X2)
    cnam2 <- dimnames(X2)[[2]]
    if (dim(X1)[2] > 0 && p2 > 1 && cnam2[1] == "(Intercept)") {
        X1 <- cbind(X2[, 1], X1)
        dimnames(X1)[[2]][1] <- "(Intercept)"
        X2 <- X2[, -1, drop = FALSE]
        x1.idx <- c(1L, x1.idx)
    }
    else if (p2 == 1 && cnam2 == "(Intercept)") {
        X1 <- X
        x1.idx <- 1:ncol(X)
        X2 <- NULL
    }
    if (length(dim(X1)) && dim(X1)[2] == 0) {
        X1 <- NULL
        x1.idx <- NULL
    }
    if (length(dim(X2)) && dim(X2)[2] == 0)
        X2 <- NULL
    fit <- if (length(weights))
        lmRob.wfit(X, Y, weights, x1 = X1, x2 = X2, x1.idx = x1.idx,
            nrep = nrep, robust.control = control, genetic.control = genetic.control)
    else lmRob.fit(X, Y, x1 = X1, x2 = X2, x1.idx = x1.idx, nrep = nrep,
        robust.control = control, genetic.control = genetic.control)
    if (is.null(fit))
        return(NULL)
    fit$terms <- Terms
    fit$call <- the.call
    x.names <- dimnames(X)[[2]]
    pasgn <- asgn
    qrx <- qr(X)
    Rk <- qrx[["rank"]]
    piv <- qrx[["pivot"]][1:Rk]
    newpos <- match(1:Rk, piv)
    if (length(x.names))
        names(newpos) <- x.names
    for (j in names(asgn)) {
        aj <- asgn[[j]]
        aj <- aj[ok <- (nj <- newpos[aj]) <= Rk]
        if (length(aj)) {
            asgn[[j]] <- aj
            pasgn[[j]] <- nj[ok]
        }
        else asgn[[j]] <- pasgn[[j]] <- NULL
    }
    effects <- X * matrix(fit$coefficients, byrow = TRUE, nrow = nrow(X),
        ncol = ncol(X))
    fit$effects <- sqrt(colSums(effects^2))
    fit$assign <- asgn
    if (model)
        fit$model <- m
    if (x)
        fit$x <- X
    if (y)
        fit$y <- Y
    attr(fit, "na.message") <- attr(m, "na.message")
    if (!is.null(attr(m, "na.action")))
        fit$na.action <- attr(m, "na.action")
    fit
}

# Data set sorted by experimental condition (stock, libxul, static)
# and then in increasing order of hard drive performance.  Discard the
# first data point in each set, which is nonrepresentative because it
# does some setup work that is cached for later runs, as well as
# priming operating-system and CPU caches.

timings.raw <- data.frame(dynamic.4200=scan('data-9.10-4200rpm-USB/stock.csv', skip=1),
                          dynamic.5400=scan('data-9.10-5400rpm/stock.csv', skip=1),
                          dynamic.7200=scan('data-9.10-7200rpm/stock.csv', skip=1),
                          dynamic.ssd=scan('data-9.10-ssd/stock.csv', skip=1),
                          libxul.4200=scan('data-9.10-4200rpm-USB/libxul.csv', skip=1),
                          libxul.5400=scan('data-9.10-5400rpm/libxul.csv', skip=1),
                          libxul.7200=scan('data-9.10-7200rpm/libxul.csv', skip=1),
                          libxul.ssd=scan('data-9.10-ssd/libxul.csv', skip=1),
                          static.4200=scan('data-9.10-4200rpm-USB/static.csv', skip=1),
                          static.5400=scan('data-9.10-5400rpm/static.csv', skip=1),
                          static.7200=scan('data-9.10-7200rpm/static.csv', skip=1),
                          static.ssd=scan('data-9.10-ssd/static.csv', skip=1)
                          )

# Reorganize the data the way lmRob and ggplot like it.
timings <- melt(timings.raw, id.vars=c(), variable_name='condition')
timings <- rename(cbind(colsplit(timings$condition, split='\\.',
                                 names=c("build", "disk")),
                        subset(timings, TRUE, "value")),
                  c(value='time'))

# Report medians for comparison
print(cast(melt(timings, measure.vars='time'), build ~ disk, fun.aggregate=median))

# Take the model
model <- lmRob(time ~ disk * build, data=timings)

options(digits=3)
print(anova(model))
print(summary(model))

# Normal q-q plot of residuals
pdf(file='qq-residuals-u.pdf', width=6.5, height=6.5)
print(plot(model, 2))
dev.off()
system('qpdf qq-residuals-u.pdf qq-residuals.pdf')

# Predictions with data
pred <- predict(model, se=TRUE)

timings$time.p <- pred$fit
timings$time.ph <- pred$fit + 2*pred$se.fit
timings$time.pl <- pred$fit - 2*pred$se.fit

pdf(file='full-prediction-u.pdf', width=6.5, height=9)
p <- ggplot(data=timings, aes(x=disk, colour=build, group=build)) +
  geom_point(aes(y=time), position=position_jitter(w=0.1, h=0), alpha=0.4) +
  geom_smooth(stat="identity", aes(y=time.p, ymin=time.pl, ymax=time.ph))
print(p)
dev.off()
system('qpdf full-prediction-u.pdf full-prediction.pdf')

# Residuals, subtracting off the effect of disk
trend <- coef(model)[c("(Intercept)","disk5400","disk7200","diskssd")]
trend[2:4] = trend[2:4] + trend[1]
names(trend) <- c("4200","5400","7200","ssd")

detrended <- ddply(timings, .(disk), function(block) {
  delta <- trend[block$disk[1]]
  return(transform(block,
                   time = time - delta,
                   time.p = time.p - delta,
                   time.ph = time.ph - delta,
                   time.pl = time.pl - delta))
})

pdf(file='bld-prediction-u.pdf', width=6.5, height=9)
p <- ggplot(data=detrended, aes(x=disk, colour=build, group=build)) +
  geom_point(aes(y=time), position=position_jitter(w=0.1, h=0), alpha=0.4) +
  geom_smooth(stat="identity", aes(y=time.p, ymin=time.pl, ymax=time.ph)) +
  ylim(-1200, 500) + ylab('Change in time due to build (milliseconds)')
print(p)
dev.off()
system('qpdf bld-prediction-u.pdf bld-prediction.pdf')

